local nametag, addon = ...
local L = LibStub("AceLocale-3.0"):GetLocale(nametag)
local small_button_image_size,large_button_image_size = 16,24
local small_font_object,large_font_object = GameFontHighlightSmall,GameFontHighlight

addon.defaults = {
	profile = {
		enable = true,
		debug = 1,
		scope = 'realm',
		--name_placement = 'header',
		gvault = true,
		size = 'small',
		give_location = true,
		remind = true,
		notimeplayed = true,
		slash_rd = true,
		slash_what = true,
	},
}

addon.options = {
	--name = filled in during OnInit
	type = 'group',
	childGroups = 'tab',
	handler = addon,    -- functions listed as strings called as addon:func
	get = "GetOption",
	set = "SetOption",
	args = {
		general = {
			name = GENERAL,
			desc = L["General options"],
			type = 'group',
			order = 10,
			args = {
				version = {
					--name = filled in during OnInit
					type = 'description',
					fontSize = "large",
					cmdHidden = true,
					width = 'full',
					order = 1,
				},
				enable = {
					name = ENABLE,
					desc = L["Use this addon"],
					type = 'toggle',
					arg  = "ToggleEnable",
					order = 5,
				},
				gvault = {
					name = L["Guild vaults"],
					desc = L["Also search guild vault contents"],
					type = 'toggle',
					order = 15,
				},
				give_location = {
					name = L["Show location"],
					desc = L["Include bag and slot #; turn off if using a 'one huge bag' inventory addon"],
					type = 'toggle',
					arg  = "UpdateOptionLocals",
					order = 17,
				},
				remind = {
					name = L["Bag# Reminders"],
					desc = L["Add '<'/'>' bag direction reminders (see Help)"],
					type = 'toggle',
					arg  = "UpdateOptionLocals",
					order = 25,
				},
				break1 = {
					name = '',
					type = 'description',
					cmdHidden = true,
					width = 'full',
					order = 50,
				},
				scope = {
					name = L["Scope"],
					desc = L["How 'wide' a search should be"],
					type = 'select',
					order = 60,
					--values = filled in below
				},
				--[==[
				name_placement = {
					name = "Name Placement",
					desc = "Where character names appear in the output",
					type = 'select',
					order = 70,
					--values = filled in below
					disabled=true,hidden=true,
				},]==]
				size = {
					name = L["Row Height"],
					type = 'select',
					arg  = "UpdateOptionLocals",
					order = 80,
					--values = filled in below
				},
				break2 = {
					name = '',
					type = 'description',
					cmdHidden = true,
					width = 'full',
					order = 90,
				},
				slash_rd = {
					name = L["Register /rd"],
					desc = L["Register /rd as a synonym for /rundown (changing requires /reload)"],
					type = 'toggle',
					order = 91,
				},
				slash_what = {
					name = L["Register /what"],
					desc = L["Register /what as a synonym for /rundown (changing requires /reload)"],
					type = 'toggle',
					order = 92,
				},
				notimeplayed = {
					name = L["Suppress /played login spam"],
					desc = L["Suppress DataStore's /played output during login"],
					type = 'toggle',
					order = 93,
				},
				-- sigh.  "-2" does not get placed in front of -1
				break3 = {
					name = '',
					type = 'description',
					cmdHidden = true,
					width = 'full',
					order = 300,
				},
				debug = {
					name = L["Debug"],
					desc = L["Metric crapton of output in chat"],
					type = 'select',
					arg  = "UpdateOptionLocals",
					order = 301,
				},
				main = {
					--name =  filled in locals
					type = 'execute',
					func = function()
						if addon.optionsFrame:IsVisible() then
							InterfaceOptionsFrameCancel:Click()
							HideUIPanel(GameMenuFrame)
						else
							LibStub("AceConfigDialog-3.0"):Close(nametag)
						end
						GameTooltip:Hide()
						addon:OnChatCommand()
					end,
					order = 302,
				},
			},
		},
		--profiles =   filled in during OnInit
		help = {
			name = HELP_LABEL,
			desc = L["Instructions"],
			type = 'group',
			order = -1,
			args = {},
		},
	},
}

addon.options.args.general.args.scope.values = {
	['realm'] = L["Current realm only"],
	['account'] = L["Cross-realm"],
	--['computer'] = [[Cross-account]],
}
--[==[
addon.options.args.general.args.name_placement.values = {
	['header'] = [[Line separator]],
	['column'] = [[Additional column]],
}]==]
addon.options.args.general.args.size.values = {
	['small'] = SMALL,
	['large'] = LARGE,
}
addon.options.args.general.args.debug.values = {
	[1] = OFF,
	[2] = [[Basic (toons & matches)]],
	[3] = [[Basic + UI]],
	[4] = [[Basic + slot contents]],
	[5] = ALL,
}
addon.options.args.help.args.main = addon.options.args.general.args.main

local options_button_open = L["Open Rundown"]
local options_button_back = L["Back To Rundown"]
addon.options.args.general.args.main.name = options_button_open

-----------------------------------------------------------------------------
-- other locals
local DS_expected_verstring = "3.2.002"  -- prefix only
local datafail_error = [[failed getting %s data for %s, skipping - error follows:  %s]]
local type, pairs, ipairs, strsplit = _G.type, _G.pairs, _G.ipairs, _G.strsplit
local vault_height
local DS -- DataStore after loading
local AceGUI = LibStub("AceGUI-3.0")
--local flib = LibStub("LibFarmbuyer")
local new, del, copy, clear
local results, searchtargets, most_recent_blockname = {}
local get_safe_datastore_function
local can_search  -- bags (incl vault), slots (equipped), char (class info etc)
local size,remind,debug,give_location

local cachefail
local cache_rescan_delay = 7
local function GetItemInfo(x)
	local name,link
	if type(x) == 'string' then
		link = x
		name = link:match('%[([^%]]+)%]|h')
		if name then return name,link end
	end
	name,link = _G.GetItemInfo(x)
	if name then return name,link end
	cachefail = true
end

local FMT = {
	BAG_plain =    "%s %d:  %s",
	BAG_remind =   "%s <%d:  %s",
	BANK_plain =   "%s %d:  %s",
	BANK_remind =  "%s >%d:  %s",
}

local l_BagSlot_fmt = L["Slot %d"]
local l_Vault_coords_fmt = L["Row %d, Column %d"]
local l_BAG = _G.BAGSLOT
local l_BANK_STORAGE = L["Bank"] -- no globalstrings entry for 'bank'?  wtf.
local l_BANK_BAG = L["Bank Bag"] -- "Bag Slot" is not a great label, don't use globalstring here
local dsbags_to_names = {
	['0'] = BACKPACK_TOOLTIP,
	['-2'] = KEYRING,
	['100'] = l_BANK_STORAGE,
	['-1'] = l_BANK_STORAGE,  -- in case DS ever uses the same indexing scheme as the rest of the world
}

-- Using "-" or "/" followed by any of these words opens the corresponding config
-- panel instead of the main window.
local slash_words = {
	["options"] = 'general', ["option"] = 'general', ["opt"] = 'general', ["config"] = 'general',
	["help"] = 'help', ["?"] = 'help',
}


-----------------------------------------------------------------------------
addon = LibStub("AceAddon-3.0"):NewAddon(addon, nametag,
		        "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")

-- Its on-disk name and the global table name are the same, which simplifies things.
function addon:DS_sanity_check (modulename)
	if not _G[modulename] then return end
	local ver = GetAddOnMetadata(modulename, "Version") or "?"
	-- force this to either true or nil
	return ver:match('^'..DS_expected_verstring) and true
end

-- Thanks to jerry for the nifty arg idea.
function addon:SetOption (info, value)
	local name = info[#info]
	self.db.profile[name] = value
	local arg = info.arg
	if arg then self[arg](self) end
end

function addon:GetOption (info)
	local name = info[#info]
	return self.db.profile[name]
end

function addon:DPrint (d,s,...)
	if d > debug then return end
	if d==3 and debug==4 then return end -- FIXME use some proper bitflags you tard
	if s:find('%',nil,true) then
		self:Print(s:format(...))
	else
		self:Print(s,...)
	end
end

function addon:OnInitialize()
	self.pretty_verstring = nametag .. " |cff30adff" .. (GetAddOnMetadata(nametag, "Version") or "?") .. "|r"
	self.db = LibStub("AceDB-3.0"):New("RundownIVDB", self.defaults, --[[Default=]]true)

	local AceDBOptions = LibStub("AceDBOptions-3.0", true)
	if AceDBOptions then
		self.options.args.profiles = AceDBOptions:GetOptionsTable(self.db)
		self.options.args.profiles.order = 200
	end

	self.options.name = GetAddOnMetadata(nametag, "Title")
	self.options.args.general.args.version.name = self.pretty_verstring
	LibStub("AceConfig-3.0"):RegisterOptionsTable(nametag, self.options)
	self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions(nametag, "Rundown")
	self.optionsFrame.okay = function() 
		addon.options.args.general.args.main.name = options_button_open
	end
	self.optionsFrame.refresh = self.optionsFrame.okay
	self.optionsFrame.cancel = self.optionsFrame.okay

	-- This only works because of the initialize->enable behavior of aceaddon.
	-- DS_Char will always be in front of us in either queue.
	-- <DS_Char New> -> <R4 New> -> <DS_Char OnInit> -> <R4 OnInit> -> <DS_Char OnEnable> -> <R4 OnEnable>
	if self.db.profile.notimeplayed then
		self.old_RTP = _G.RequestTimePlayed
		_G.RequestTimePlayed = function() end
	end

	self:SetEnabledState(self.db.profile.enable)
	self:UpdateOptionLocals()

	GuildBankFrame_LoadUI()
	vault_height = assert(NUM_SLOTS_PER_GUILDBANK_GROUP/2, "failed to load GuildBankFrameUI")

	self.OnInitialize = nil
end


function addon:OnEnable()
	if self.old_RTP then
		_G.RequestTimePlayed = self.old_RTP
		self.old_RTP = nil
	end
	self:RegisterEvent("PLAYER_LOGOUT")
	if not self.db.profile.enable then
		return self:Disable()
	end

	can_search = {
		-- char modules is not strictly speaking part of what gets searched
		char = self:DS_sanity_check('DataStore_Characters'),
		bags = self:DS_sanity_check('DataStore_Containers'),
		slots = self:DS_sanity_check('DataStore_Inventory'),
	}
	local sane
	for k,v in next, can_search do
		sane = sane or v
	end
	if not sane then
		self:Print(self.error_exit)
		return
	end

	-- sigh
	if can_search.bags then
		self:RegisterEvent("GUILDBANKBAGSLOTS_CHANGED")
	end

	self.error_exit = nil
	self.DS_sanity_check = nil
	DS = _G.DataStore
	self:RegisterChatCommand("rundown", "OnChatCommand")
	-- Instead of creating multiple copies of the same tiny wrapper, abuse knowledge
	-- of implementation to get a reference to the created wrapper, and just reuse
	-- that as the handler.
	local slashfunc = _G.SlashCmdList[ LibStub("AceConsole-3.0").commands["rundown"] ]
	assert(type(slashfunc)=='function')
	if self.db.profile.slash_rd then
		self:RegisterChatCommand("rd", slashfunc)
	end
	if self.db.profile.slash_what then
		self:RegisterChatCommand("what", slashfunc)
	end
	table.insert(UISpecialFrames,nametag)
end

function addon:OnDisable()
end

function addon:ToggleEnable()
	if self.db.profile.enable then
		self:Enable()
	else
		self:Disable()
	end
end

-- locals are good
function addon:UpdateOptionLocals()
	size = self.db.profile.size
	remind = self.db.profile.remind
	debug = self.db.profile.debug
	give_location = self.db.profile.give_location
end

function addon:OnChatCommand (input)
	if input then
		input = input:trim()
		local p,word = input:sub(1,1), input:sub(2)
		if p == '-' or p == '/' then
			word = slash_words[word]
			if word then
				LibStub("AceConfigDialog-3.0"):Open(nametag)
				LibStub("AceConfigDialog-3.0"):SelectGroup(nametag,word)
				return
			end
		end
	end

	if (not self.display) or (not self.display:IsShown()) then
		self:BuildWindow()
	end

	if (not input) or (#input == 0) then
		self.display:SetText(L["Type or drag an item here"])
		self.display.editbox:HighlightText()
		return
	end
	self:DoSearch(input)
end

function addon:PLAYER_LOGOUT()
	self.db.profile.debug = nil
end

-- Tables for holding results are reused until the player gets into a fight.
function addon:PLAYER_REGEN_DISABLED()
	self:UnregisterEvent("PLAYER_REGEN_DISABLED")
	if self.display and self.display:IsShown() then
		self.display:Hide()
	end
	most_recent_blockname = nil
	wipe(results)
	clear()
end

-- This is making up for a mild deficiency in DataStore_Containers.  Lots of
-- hardcoded stuff here, see DataStore_Containers.lua:ScanGuildBankInfo().
-- The excruciatingly defensive coding here is because the event can fire at
-- zone transitions (!), and the Get* calls can return nil just then.  Sigh.
do
	local _guildname = GetGuildInfo("player")
	local _realm = GetRealmName()
	function addon:GUILDBANKBAGSLOTS_CHANGED()
		if not _guildname then
			_guildname = GetGuildInfo("player")
		end
		if not _realm then
			_realm = GetRealmName()
		end
		local tabID = GetCurrentGuildBankTab()
		local key = ("%s.%s.%s"):format('Default', _realm, _guildname)
		local t = _G.DataStore_Containers.db.global.Guilds[key][tabID]
		t.name,t.icon = GetGuildBankTabInfo(tabID)
	end
end

function addon:find (needle, haystack)
	-- go go gadget super-defensive coding
	if needle == nil or haystack == nil then return end
	haystack = tostring(haystack):lower()

	-- hardcode to plain
	return haystack:find(needle,nil,true)
end


-----------------------------------------------------------------------------
function addon:add_to_results (blockname, whence, itemlink, ...)
	self:DPrint(2, "***", blockname, whence, itemlink, ...)

	local entry
	if blockname ~= most_recent_blockname then
		entry = new()
		entry.name = blockname
		--entry.whence = whence -- needed?
		table.insert(results,entry)
		most_recent_blockname = blockname
	end

	-- last two may be nil
	local cID, slot, count = ...
	cID = tostring(cID)
	slot = slot and tonumber(slot)

	entry = new()
	entry.itemlink = itemlink
	entry.whence = whence
	entry.count = count and tonumber(count)

	-- I was really expecting more here.  Possbly merge with :DoSearch.

	if whence == 'slots' then
		-- nothing else to do...?

	elseif whence == 'vault' then
		entry.tabname, entry.tabicon = select(4,...)
		if slot and give_location then
			--[===[
			A vault tab is 7 rows tall and 7 "columns" wide, with each "column"
			being two actual columns of buttons:
				--Col1-- --Col2--   Slot  1  ==>  C1B1
				|B1||B8| |B1||B8|   Slot  2  ==>  C1B2
				|B2||B9| |B2||B9|   Slot  8  ==>  C1B8
				|B3||B10   ...      Slot 15  ==>  C2B1
				|B4||B11
				|B5||B12
				|B6||B13
				|B7||B14
			]===]
			local index, col, row
			index = slot % NUM_SLOTS_PER_GUILDBANK_GROUP -- =14
			if index == 0 then
				index = NUM_SLOTS_PER_GUILDBANK_GROUP
			end
			col = 2 * math.ceil((slot-0.5)/NUM_SLOTS_PER_GUILDBANK_GROUP)

			-- row = index % vault_height (and 0->vault_height)
			if index <= vault_height then
				row = index
				col = col - 1
			else
				row = index - vault_height
			end

			entry.slot = l_Vault_coords_fmt:format(row,col)
		end

	elseif whence == 'bags' then
		entry.baglink, entry.bagicon = select(4,...)
		if cID:sub(1,3) == 'Bag' then
			local dsbag,s_dsbag
			s_dsbag = cID:sub(4); dsbag = tonumber(s_dsbag)
			if dsbag > 0 and dsbag <= NUM_BAG_SLOTS then
				-- 'Bag 2: [blah]'
				entry.baglink = give_location
				                and FMT[remind and 'BAG_remind' or 'BAG_plain']:format(BAGSLOT,dsbag,entry.baglink)
				                or l_BAG

			elseif dsbag > NUM_BAG_SLOTS and dsbag <= NUM_BAG_SLOTS+NUM_BANKBAGSLOTS then
				-- 'Bank Bag 6: [blah]'
				entry.baglink = give_location
				                and FMT[remind and 'BANK_remind' or 'BANK_plain']:format(l_BANK_BAG,dsbag-NUM_BAG_SLOTS,entry.baglink)
				                or l_BANK_BAG

			else
				entry.baglink = dsbags_to_names[s_dsbag]

			end
		else
			-- I don't think we can usefully figure out the container ID, if datastore hasn't already done so.
		end
		if slot and give_location then
			entry.slot = l_BagSlot_fmt:format(slot)
		end

	end

	table.insert(results,entry)
end


-- In all these following functions, the CHAR parameter is a DS character object.

-- DataStore information can be remarkably patchy and full of gaps.  If toon
-- X has been seen by (e.g.) DS_Containers, and then DS_Characters is installed,
-- but X has not yet logged on, then toon Y will get X as part of the loops
-- below -- except that DS_Characters methods will fail with nil errors due to
-- the lack of checking.  Possibly this never happens to normal users.  Maybe
-- all the time.  Who knows.  To be fair, I'm not sure what I would do either.
--
-- As an attempt to avoid this, we prod modules with one of their supplied
-- methods, as it becomes convenient.  If any of them fail, we exclude that
-- character from the loops.
-- XXX should notify user to login to that toon & examine bank
do
	local xpcall = _G.xpcall
	local function err(...) return ... end
	local safe_functions = {}
	function get_safe_datastore_function(ds_func_name)
		if not safe_functions[ds_func_name] then
			safe_functions[ds_func_name] = function (char)
				local function f() return DS[ds_func_name](DS,char) end
				return xpcall (f, err)
			end
		end
		return safe_functions[ds_func_name]
	end
end

-- Returns a name appropriate for displaying to the user, and also as a
-- unique label for subsequent searches.  Returns nil if this character
-- should be avoided.
function addon:GetToonDisplayName (fallback, char, realm)
	local ret, nstring
	local scope = self.db.profile.scope

	if can_search.char then
		ret, nstring = get_safe_datastore_function('GetColoredCharacterName')(char)
		if not ret then
			self:DPrint(2, datafail_error, '<Character module>', tostring(char), nstring)
			return
		end
		if scope == 'realm' then           -- name is sufficient
			nstring = nstring .. '|r'
		elseif scope == 'account' then      -- use name-realm
			nstring = nstring .. ' - ' .. realm .. '|r'
		elseif scope == 'computer' then    -- for now, only name-realm is useful
			nstring = nstring .. ' - ' .. realm .. '|r'
		end
	else
		nstring = fallback -- really? that's all we can manage? crumbs.
	end

	return nstring
end
function addon:GetGuildDisplayName (guildname, realmname)
	-- Less fragile than for characters.  Multiple string concats in a single
	-- statement are faster than string.format when the variables are already
	-- of string type.
	if realmname then
		return ORANGE_FONT_COLOR_CODE .. '<' .. guildname .. '> - ' .. realmname .. '|r'
	else
		return ORANGE_FONT_COLOR_CODE .. '<' .. guildname .. '>|r'
	end
end

function addon:SearchCharacterSlots (blockname, char, texts)
	self:DPrint(2, ">>> starting :SearchCharacterSlots", blockname)
	local ret, inv = get_safe_datastore_function('GetInventory')(char)
	if not ret then
		self:DPrint(2, datafail_error, '<Inventory module>', tostring(char), inv)
		return
	end

	-- These entries might be itemlinks (if DS thinks they're enchanted) or plain
	-- itemIDs.  Either can safely be fed to GetItemInfo, which will fill out any
	-- missing bits and give a complete link.
	for i,text in ipairs(texts) do
		-- ipairs(inv) stops on the first empty slot, so we go a little more clunky
		for slot = 1, INVSLOT_LAST_EQUIPPED do while true do
			local item = inv[slot]
			if not item then break end   -- fake a continue
			local itemname, itemlink = GetItemInfo(item)
			self:DPrint(4, "*inv", text, slot, item, itemlink)
			if self:find (text, itemname) then
				self:add_to_results (blockname, 'slots', itemlink)
			end
			break
		end end
	end

	self:DPrint(2, "<<< done with :SearchCharacterSlots", blockname)
end


function addon:SearchContainer (blockname, text, containerID, container, is_vault)
	self:DPrint(4, "*cnt", containerID)
	if type(container.ids) ~= 'table' then return end  -- empty bag
	if type(container.size) ~= 'number' then return end  -- empty guild vault tab
	for slot = 1, container.size do while true do
		local itemid, itemname, itemlink
		itemid = container.ids[slot]
		if not itemid then break end  -- empty slot
		itemlink = container.links and container.links[slot] 
		-- itemlink might be nil if there's nothing special about that item
		itemname, itemlink = GetItemInfo(itemlink or itemid)
		self:DPrint(4, "*cntslot", slot, itemid, itemname, itemlink)
		if self:find (text, itemname) then
			if is_vault then
				self:add_to_results (blockname, 'vault', itemlink,
					containerID, slot, container.counts and container.counts[slot],
					container.name, container.icon)
			else
				self:add_to_results (blockname, 'bags', itemlink,
					containerID, slot, container.counts and container.counts[slot],
					container.link, container.icon)
			end
		end
		break
	end end
end


function addon:SearchGuildVault (blockname, guild_as_char, texts)
	self:DPrint(2, ">>> starting :SearchGuildVault", blockname, #guild_as_char)

	-- DataStore doesn't check nor store the number of tabs for a given vault.
	for i,text in ipairs(texts) do
		for tabid = 1, #guild_as_char do
			local tab = DS:GetGuildBankTab(guild_as_char,tabid)
			if tab.name then -- bit of a hack here
				self:SearchContainer (blockname, text, tabid, tab, true)
			end
		end
	end

	self:DPrint(2, "<<< done with :SearchGuildVault", blockname)
end


function addon:SearchCharacterBags (blockname, char, texts)
	self:DPrint(2, ">>> starting :SearchCharacterBags", blockname)
	local ret, all = get_safe_datastore_function('GetContainers')(char)
	if not ret then
		self:DPrint(2, datafail_error, '<Containers module>', tostring(char), all)
		return
	end

	for i,text in ipairs(texts) do
		for bagid, bagdata in pairs(all) do
			self:SearchContainer (blockname, text, bagid, bagdata, false)
		end
	end

	self:DPrint(2, "<<< done with :SearchCharacterBags", blockname)
end


function addon:SearchCharacter (blockname, char, texts)
	self:DPrint(2, ">> starting :SearchCharacter", blockname)

	if can_search.slots then
		self:SearchCharacterSlots (blockname, char, texts)
	end

	if can_search.bags then
		self:SearchCharacterBags (blockname, char, texts)
	end

	self:DPrint(2, "<< done with :SearchCharacter", blockname)
end


function addon:Search (texts)
	self:DPrint(2, "> starting :Search")
	cachefail = nil
	most_recent_blockname = nil
	while #results > 0 do
		del(table.remove(results))
	end

	local scope = self.db.profile.scope
	assert(scope=='realm' or scope=='account' or scope=='computer', L["ERROR:  Scope settings are broken.  (How did you manage that?)  You should open the options panel and reset the current profile to default settings."])

	local do_vaults = self.db.profile.gvault

	if scope == 'realm' then
		self:DPrint(2, "Scope:  cross-toons")
		for name, character in pairs(DS:GetCharacters()) do
			local blockname = self:GetToonDisplayName(name, character)
			if blockname then
				self:SearchCharacter (blockname, character, texts)
			end
		end
		if do_vaults then
			local ret, all = get_safe_datastore_function('GetGuilds')()
			if all then
				for name, guild in pairs(all) do
					self:SearchGuildVault (self:GetGuildDisplayName(name), guild, texts)
				end
			end
		end

	elseif scope == 'account' then
		self:DPrint(2, "Scope:  cross-realm")
		for realm in pairs(DS:GetRealms()) do
			for name, character in pairs(DS:GetCharacters(realm)) do
				local blockname = self:GetToonDisplayName(name, character, realm)
				if blockname then
					self:SearchCharacter (blockname, character, texts)
				end
			end
			if do_vaults then
				local ret, all = get_safe_datastore_function('GetGuilds')(realm)
				if all then
					for name, guild in pairs(all) do
						self:SearchGuildVault (self:GetGuildDisplayName(name,realm), guild, texts)
					end
				end
			end
		end

	--[====[
	elseif scope == 'computer' then
		-- This is largely useless until we (and DS) figure out a way of labelling an account.
		self:DPrint(2, "Scope:  cross-account")
		for account in pairs(DS:GetAccounts()) do
			for realm in pairs(DS:GetRealms(account)) do
				for name, character in pairs(DS:GetCharacters(realm,account)) do
					local blockname = self:GetToonDisplayName(name, character, realm)
					if blockname then
						self:SearchCharacter (blockname, character, texts)
					end
				end
				if do_vaults then
					local ret, all = get_safe_datastore_function('GetGuilds')(realm,account)
					if all then
						for name, guild in pairs(all) do
							self:SearchGuildVault (self:GetGuildDisplayName(name,realm), guild, texts)
						end
					end
				end
			end
		end]====]
	end

	self:DPrint(2, "< done with :Search")
end


-----------------------------------------------------------------------------
local function BlocklineHeading_OnRelease (widget)
	widget.label:ClearAllPoints()
	widget.label:SetPoint("TOP")
	widget.label:SetPoint("BOTTOM")
	widget.OnRelease = nil
end

local function mkblockline (text)
	local line = AceGUI:Create("Heading")
	line:SetText(text)
	line:SetFullWidth(true)
	line:SetHeight(30)
	-- No API exists for this, and it takes forever getting them to add one.
	-- Just do it ourselves and put things back when we're done.
	-- Also, :SetJustifyH("LEFT") doesn't do a #$%#$%#%#% thing no matter
	-- what actions it's combined with.
	line.OnRelease = BlocklineHeading_OnRelease
	line.label:ClearAllPoints()
	line.label:SetPoint("LEFT", 25, 0)
	return line
end

local function LineLabel_OnClick (widget)
	if IsModifiedClick("CHATLINK") and ChatEdit_GetActiveWindow() then
		-- totally cheating here:  get the first widget in the row and pull out its text
		ChatEdit_InsertLink(widget.parent.children[1].label:GetText())
	end
end
local function LineLabel_OnEnter (widget)
	if widget:GetUserData('isitem') then
		GameTooltip:SetOwner (widget.frame, "ANCHOR_LEFT", -20, 0)
		GameTooltip:SetHyperlink (widget.label:GetText())
	end
	--widget.parent:SetHighlight(1.0,0.9,0.0,0.5)  -- lib-st defaults, yellowish
	widget.parent:SetHighlight (0.3, 0.3, 0.3, 0.5)
end
local function LineLabel_OnLeave (widget)
	widget.parent:SetHighlight(0,0,0,0)
	GameTooltip:Hide()
end

local function mkitemlabel (itemlink, count)
	local itemlabel = AceGUI:Create("InteractiveCountingLabel")
	itemlabel:SetUserData('isitem',true)
	if size == 'small' then
		--itemlabel.count:SetFontObject(NumberFontNormalSmall)
		itemlabel:SetFontObject(small_font_object)
		itemlabel:SetImageSize(small_button_image_size,small_button_image_size)
		if count and count > 1 then
			itemlabel:SetText(itemlink .. ' x' .. count)
		else
			itemlabel:SetText(itemlink)
		end
	elseif size == 'large' then
		itemlabel.count:SetFontObject(NumberFontNormal)
		itemlabel:SetFontObject(large_font_object)
		itemlabel:SetImageSize(large_button_image_size,large_button_image_size)
		itemlabel:SetText(itemlink)
		itemlabel:SetCountText(count)
	end
	itemlabel:SetRelativeWidth(0.4)
	itemlabel:SetCallback("OnClick", LineLabel_OnClick)
	itemlabel:SetCallback("OnEnter", LineLabel_OnEnter)
	itemlabel:SetCallback("OnLeave", LineLabel_OnLeave)
	local id = itemlink:match("item:(%d+)")
	itemlabel:SetImage(GetItemIcon(id))
	return itemlabel
end

local function mklocationlabel (text, iconpath)
	local loclabel = AceGUI:Create("InteractiveCountingLabel")
	if size == 'small' then
		loclabel:SetFontObject(small_font_object)
		loclabel:SetImageSize(small_button_image_size,small_button_image_size)
	elseif size == 'large' then
		loclabel:SetFontObject(large_font_object)
		loclabel:SetImageSize(large_button_image_size,large_button_image_size)
	end
	loclabel:SetRelativeWidth(0.4)
	loclabel:SetCallback("OnClick", LineLabel_OnClick)
	loclabel:SetCallback("OnEnter", LineLabel_OnEnter)
	loclabel:SetCallback("OnLeave", LineLabel_OnLeave)
	loclabel:SetText(text)
	loclabel:SetImage(iconpath)
	return loclabel
end

local function mkslotline (itemlink)
	local itemlabel = mkitemlabel(itemlink)
	local eq = mklocationlabel(CURRENTLY_EQUIPPED)
	eq:SetRelativeWidth(0.59)  -- fixup

	local line = AceGUI:Create("InteractiveSimpleGroup")
	line:SetLayout("Flow")
	line:SetFullWidth(true)
	line:AddChildren (itemlabel,eq)
	return line
end

local function mkbagsline (itemlink, bagtext, bagicon, slot, count)
	local itemlabel = mkitemlabel(itemlink,count)
	local bag = mklocationlabel(bagtext,bagicon)
	local slotw = mklocationlabel(slot)
	slotw:SetRelativeWidth(0.19)  -- fixup

	local line = AceGUI:Create("InteractiveSimpleGroup")
	line:SetLayout("Flow")
	line:SetFullWidth(true)
	line:AddChildren (itemlabel,bag,slotw)
	return line
end

--[[
local function mkvaultline (itemlink, bagtext, bagicon)
	local itemlabel = mkitemlabel(itemlink)
	local bag = mklocationlabel(bagtext,bagicon)
	local line = AceGUI:Create("SimpleGroup")
	line:SetLayout("Flow")
	line:SetFullWidth(true)
	line:AddChildren (itemlabel,bag)
	return line
end]]

local function SpecialBottomButtonClicked (display, ev, button, key)
	display:Hide()
	addon.options.args.general.args.main.name = options_button_back
	LibStub("AceConfigDialog-3.0"):Open(nametag)
	LibStub("AceConfigDialog-3.0"):SelectGroup(nametag,key)
end

local function EditBox_OnFocus (display, event)
	display.editbox:HighlightText(0, event == "OnFocusGained" and -1 or 0)
end

local function EditBox_OnEnterPressed (display, event, text)
	addon:DoSearch(text)
end

function addon:DoSearch (text, redoing_p)
	assert(self.display and self.display:IsShown(), "searching without window")
	self.scroll:ReleaseChildren()

	if redoing_p then
		-- searchtargets preserved, ignore 'text'
	else
		text = text:trim()
		if #text < 1 then
			self:DPrint(2, "DoSearch given an empty string!")
			return
		end
		searchtargets = new(strsplit(';',text))
		local i = #searchtargets
		while i > 0 do
			local txt = searchtargets[i]
			if not txt then
				table.remove(searchtargets,i)
				i = #searchtargets  -- meh
			elseif #txt < 3 then
				self:Printf("%s: %s", txt, L["Search string is too short!"])
				table.remove(searchtargets,i)
				i = #searchtargets  -- meh
			else
				-- if they've linked an item to search, strip out the possibly-unique
				-- parts and just go with the item name
				searchtargets[i] = (txt:match("item:%d.*%[([^]]+)%]") or txt):lower()
				i = i - 1
			end
		end
		if #searchtargets < 1 then
			del(searchtargets)
			return
		end
	end

	-- feedback to the user is good
	self.display:SetText(table.concat(searchtargets, "; "))

	self:Search(searchtargets)

	self:DPrint(3, "BuildWindow d=%s, d.frame=%s", tostring(self.display), tostring(self.display.frame))
	self:DPrint(3, "BuildWindow s=%s, s.frame=%s, s.sf=%s, s.c=%s",
		tostring(self.scroll), tostring(self.scroll.frame), tostring(self.scroll.scrollframe), tostring(self.scroll.content))

	self.display:PauseLayout()
	if cachefail then
		if redoing_p then
			-- This might both (a) never happen, and (b) not be true, since the
			-- cache does not get written to disk.
			self:Print(L["There are many missing items in your game cache.  You should revisit the banks for your alts."])
		else
			-- Listening for GET_ITEM_INFO_RECEIVED doesn't actually help us here.
			self.scroll:AddChild(mkblockline(L["Item cache error; the results below may be incomplete.  Automatically re-scanning in %d seconds..."]:format(cache_rescan_delay)))
			self:ScheduleTimer ("ReDoSearch", cache_rescan_delay)
		end
	else
		del(searchtargets)
	end
	if #results > 0 then
		for i,result in ipairs(results) do  -- dispatch table would be overkill
			local line
			if result.name then
				line = mkblockline (result.name)

			elseif result.whence == 'slots' then
				line = mkslotline (result.itemlink)

			elseif result.whence == 'bags' then
				line = mkbagsline (result.itemlink, result.baglink, result.bagicon, result.slot, result.count)

			elseif result.whence == 'vault' then
				line = mkbagsline (result.itemlink, result.tabname, result.tabicon, result.slot, result.count)
			end

			self:DPrint(3, "Adding line %s", tostring(line.frame))
			self.scroll:AddChild(line)
		end
	else
		self:DPrint(3, BROWSE_NO_RESULTS)
		self.scroll:AddChild(mkblockline(BROWSE_NO_RESULTS))
	end

	self.display:ResumeLayout()
	self.scroll:SetScroll(0)
end


function addon:ReDoSearch()
	if (not self.display) or (not self.display:IsShown()) then
		self:BuildWindow()
	end
	self:DoSearch(nil,true)
end


function addon:BuildWindow()
	if self.display then
		self.display:Hide()
	end
	local display = AceGUI:Create("FrameWithInput")
	display:SetTitle(self.pretty_verstring)
	display:SetLayout("Fill")
	display:SetStatusTable{width=800}
	display:SetCallback("OnClose", function(_d)
		_G[nametag] = nil
		self.display = nil
		AceGUI:Release(_d)
		collectgarbage()
	end)
	self.display = display
	_G[nametag] = display  -- for Esc closing

	display:SetCallback("OnFocusGained", EditBox_OnFocus)
	display:SetCallback("OnFocusLost", EditBox_OnFocus)
	display:SetCallback("OnClick", SpecialBottomButtonClicked)
	display:SetCallback("OnEnterPressed", EditBox_OnEnterPressed)

	-- keys here match entries in addon.options.args
	display:AddButton ("general", L["Options"])  -- GAMEOPTIONS_MENU, maybe
	display:AddButton ("help", HELP_LABEL)

	local scroll = AceGUI:Create("ScrollFrame")
	scroll:SetLayout("List")
	self.scroll = scroll

	display:AddChild(scroll)
	display:ApplyStatus()
	display:SetFocus()
	display:Show()

	self:RegisterEvent("PLAYER_REGEN_DISABLED")
	return display
end


-- Recycling functions yoinked from AceConfigDialog
do
	local pool = setmetatable({},{__mode="k"})
	function clear()
		wipe(pool)
	end
	function new(...)  -- slightly modified variant, takes optional initializers
		local t = next(pool)
		if t then
			pool[t] = nil
			for i = 1, select('#',...) do
				t[i] = select(i,...)
			end
			return t
		else
			return {...}
		end
	end
	function copy(t)
		local c = new()
		for k, v in pairs(t) do
			c[k] = v
		end
		return c
	end
	function del(t)
		wipe(t)
		pool[t] = true
	end
end


addon.error_exit = "ERROR:  Either none of the required DataStore library addons are loaded, or they are an unexpected version.  You need at least one (Inventory, Containers) to search.  If you think they're installed, please double-check the version requirements on *THIS* addon's project page."

addon.options.args.help.args.genH = {
	type = 'header', width = 'full', order = 1, name = [[How To Use]]}
addon.options.args.help.args.genT = {
	type = 'description', width = 'full', order = 2, name = ([===[
Type /rundown, optionally followed by a word/phrase or an item link, to open
the window and run a search.  Only toons which have logged on and visited the
bank will be searched.  You can search for multiple phrases at once by separating
them with a semicolon.

/rundown of the boar    [finds all random-type boar gear]

/rundown peaceb; silverl   [finds all Peacebloom and Silverleaf but leaves out things
like "Peacemaker" and "Silvermoon"]

Item names containing the string will be shown, along with their location.
If the item is in a bank or bag, its slot number will also be shown on the far
right side.  Slots start at 1 in the top-left corner of a bag, and count across the row
before going to the next row down.

If there are multiple items in a stack, the stack count will be shown on the
item icon or next to the item link, depending on the row size.
]===]):gsub("([^\n])\n([^\n])", "%1 %2")}


addon.options.args.help.args.optH = {
	type = 'header', width = 'full', order = 11, name = [[Options]]}
addon.options.args.help.args.optT = {
	type = 'description', width = 'full', order = 12, name = ([===[
Rundown supports standard Ace3 profiles, but the profiles only control the
options shown here; the item data is shared across all profiles.

The scope setting controls whether your searches are limited to toons on the current realm.
If you use cross-realm searching, then the realm name will be included in the character/guild
names shown on the display.

Bag numbering reminders are included due to an oddity in how Blizzard assigns
numbers to bags.  Purchased bag slots in your bank are numbered from left to right,
and the reminder includes a ">" sign.  Equipped bags next to your backpack are
numbered from right to left, and the reminder includes a "<" sign.

If you use a slash command followed by any of "-option", "-options", "-opt", or "-config",
then you'll get the options window instead of the normal one.  Ditto for "-?" and "-help" for
the instructions window.  A slash can be used in place of the dash. 
]===]):gsub("([^\n])\n([^\n])", "%1 %2")}


addon.options.args.help.args.histH = {
	type = 'header', width = 'full', order = 21, name = [[Episode... 4?]]}
addon.options.args.help.args.histT = {
	type = 'description', width = 'full', order = 22, name = ([===[
The original KC_Rundown/KC_Items was written by Kaelten back in mid-2005.  It used
the original Ace libraries (today called Ace1).  To the best of my knowledge it was
not formally updated for Ace2, so I kludged around the changes in my local copy.
KC_Rundown continued to work through Burning Crusade and Wrath, increasingly showing
cracks in the armor.

It finally stopped working altogether in Cataclysm.  Rather than trying to find a way
to make it (and Ace1!) work under the 4.x client, I wrote Rundown from scratch.  The
"Episode 4" reflects this history; the "New Hope" reflects my hope that eventually
somebody more competent than me will take over maintanence of this project.  :-)
]===]):gsub("([^\n])\n([^\n])", "%1 %2")}

-- vim:noet
